// WinCE.cpp
// (c) 2000 Sin
//
// This file includes most neccessary initialization for Snes9x for the Dreamcast
// All Windows CE and DirectDraw initialization takes place in this file.

/* Snes9x homepage: www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#define NAME                TEXT("Sintendo")
#define TITLE               TEXT("Sintendo")

// --------- //
//  Globals  //
// --------- //

#include "global.h"
#include <windows.h>
#include <ddraw.h>
#include <d3d.h>
#include "resource.h"
#include "GUI.h"
#include "snes9x/port.h"
#include "bitmap.h"
#include "Direct3D.h"

LPDIRECTDRAW4				gpdd_directdraw = NULL;
LPDIRECTDRAWSURFACE4		gdds_screen = NULL;
LPDIRECTDRAWSURFACE4		gdds_back = NULL;
LPDIRECT3DDEVICE3			gdev_device;
LPDIRECT3D3					gd3d_direct3d;
BOOL                        gb_enabled = FALSE;
HWND						ghwd_window = NULL;

HINSTANCE					gh_inst;

// ------------ //
//  Prototypes  //
// ------------ //

void DestroyAll(void);
static HRESULT InitFail(HWND hWnd, HRESULT hr, LPCTSTR error);
long FAR PASCAL WindowProc(HWND, UINT, WPARAM, LPARAM);
HRESULT InitApp(HINSTANCE hInstance, int nCmdShow);

// ------------------------------------------------- //
//  DestroyAll()                                     //
//  Gets rid of all memory allocated before we quit  //
// ------------------------------------------------- //
void DestroyAll(void) {

	if(gdds_screen != NULL) {
		gdds_screen->Release();
		gdds_screen = NULL;
	}

	gpdd_directdraw->RestoreDisplayMode();

	if(gpdd_directdraw != NULL) {
		gpdd_directdraw->Release();
		gpdd_directdraw = NULL;
	}

//	DestroyWindow(ghwd_window);

}

// ----------------------------------------------------------------------------------------- //
//  InitFail()                                                                               //
//  Hopefully this will display useful information in the even of an initialization failure  //
// ----------------------------------------------------------------------------------------- //

// This is just a large switch statement to deal with all those DirectDraw HRESULT values
// I cut and paste this code into every DirectDraw project I've ever worked on :)
// Use it freely if you want, it's quite useful.
// Had to change it a little to support UNICODE strings.
// Don't ask about all the breaks, guess I'm just to lazy to take them out.
extern LPCTSTR ErrorMsg(int iError) {
	switch(iError) {

	case DD_OK:
		return TEXT("The request completed successfully."); break;
	case DDERR_ALREADYINITIALIZED:
		return TEXT("The object has already been initialized."); break;
	case DDERR_BLTFASTCANTCLIP:
		return TEXT("A DirectDrawClipper object is attached to a source surface that has passed into a call to the IDirectDrawSurface7::BltFast method."); break;
	case DDERR_CANNOTATTACHSURFACE:
		return TEXT("A surface cannot be attached to another requested surface."); break;
	case DDERR_CANNOTDETACHSURFACE:
		return TEXT("A surface cannot be detached from another requested surface."); break; 
	case DDERR_CANTCREATEDC:
		return TEXT("Windows cannot create any more device contexts (DCs), or a DC has requested a palette-indexed surface when the surface had no palette and the display mode was not palette-indexed (in this case DirectDraw cannot select a proper palette into the DC)."); break;
	case DDERR_CANTDUPLICATE:
		return TEXT("Primary and 3-D surfaces, or surfaces that are implicitly created, cannot be duplicated."); break;
	case DDERR_CANTLOCKSURFACE:
		return TEXT("Access to this surface is refused because an attempt was made to lock the primary surface without DCI support."); break;
	case DDERR_CANTPAGELOCK:
		return TEXT("An attempt to page-lock a surface failed. Page lock does not work on a display-memory surface or an emulated primary surface."); break;
	case DDERR_CANTPAGEUNLOCK:
		return TEXT("An attempt to page-unlock a surface failed. Page unlock does not work on a display-memory surface or an emulated primary surface."); break;
	case DDERR_CLIPPERISUSINGHWND:
		return TEXT("An attempt was made to set a clip list for a DirectDrawClipper object that is already monitoring a window handle."); break;
	case DDERR_COLORKEYNOTSET:
		return TEXT("No source color key is specified for this operation."); break;
	case DDERR_CURRENTLYNOTAVAIL:
		return TEXT("No support is currently available."); break;
	case DDERR_DCALREADYCREATED:
		return TEXT("A device context (DC) has already been returned for this surface. Only one DC can be retrieved for each surface."); break;
	case DDERR_DEVICEDOESNTOWNSURFACE:
		return TEXT("Surfaces created by one DirectDraw device cannot be used directly by another DirectDraw device."); break;
	case DDERR_DIRECTDRAWALREADYCREATED:
		return TEXT("A DirectDraw object representing this driver has already been created for this process."); break;
	case DDERR_EXCEPTION:
		return TEXT("An exception was encountered while performing the requested operation."); break;
	case DDERR_EXCLUSIVEMODEALREADYSET:
		return TEXT("An attempt was made to set the cooperative level when it was already set to exclusive."); break;
	case DDERR_EXPIRED:
		return TEXT("The data has expired and is therefore no longer valid."); break;
	case DDERR_GENERIC:
		return TEXT("There is an undefined error condition."); break;
	case DDERR_HEIGHTALIGN:
		return TEXT("The height of the provided rectangle is not a multiple of the required alignment."); break;
	case DDERR_HWNDALREADYSET:
		return TEXT("The DirectDraw cooperative-level window handle has already been set. It cannot be reset while the process has surfaces or palettes created."); break;
	case DDERR_HWNDSUBCLASSED:
		return TEXT("DirectDraw is prevented from restoring state because the DirectDraw cooperative-level window handle has been subclassed."); break;
	case DDERR_IMPLICITLYCREATED:
		return TEXT("The surface cannot be restored because it is an implicitly created surface."); break;
	case DDERR_INCOMPATIBLEPRIMARY:
		return TEXT("The primary surface creation request does not match the existing primary surface."); break;
	case DDERR_INVALIDCAPS:
		return TEXT("One or more of the capability bits passed to the callback function are incorrect."); break;
	case DDERR_INVALIDCLIPLIST:
		return TEXT("DirectDraw does not support the provided clip list."); break;
	case DDERR_INVALIDDIRECTDRAWGUID:
		return TEXT("The globally unique identifier (GUID) passed to the DirectDrawCreate function is not a valid DirectDraw driver identifier."); break;
	case DDERR_INVALIDMODE:
		return TEXT("DirectDraw does not support the requested mode."); break;
	case DDERR_INVALIDOBJECT:
		return TEXT("DirectDraw received a pointer that was an invalid DirectDraw object."); break;
	case DDERR_INVALIDPARAMS:
		return TEXT("One or more of the parameters passed to the method are incorrect."); break;
	case DDERR_INVALIDPIXELFORMAT:
		return TEXT("The pixel format was invalid as specified."); break;
	case DDERR_INVALIDPOSITION:
		return TEXT("The position of the overlay on the destination is no longer legal."); break;
	case DDERR_INVALIDRECT:
		return TEXT("The provided rectangle was invalid."); break;
	case DDERR_INVALIDSTREAM:
		return TEXT("The specified stream contains invalid data."); break;
	case DDERR_INVALIDSURFACETYPE:
		return TEXT("The surface was of the wrong type."); break;
	case DDERR_LOCKEDSURFACES:
		return TEXT("One or more surfaces are locked, causing the failure of the requested operation."); break;
	case DDERR_MOREDATA:
		return TEXT("There is more data available than the specified buffer size can hold."); break;
	case DDERR_NO3D:
		return TEXT("No 3-D hardware or emulation is present."); break;
	case DDERR_NOALPHAHW:
		return TEXT("No alpha-acceleration hardware is present or available, causing the failure of the requested operation."); break;
	case DDERR_NOBLTHW:
		return TEXT("No blitter hardware is present."); break;
	case DDERR_NOCLIPLIST:
		return TEXT("No clip list is available."); break;
	case DDERR_NOCLIPPERATTACHED:
		return TEXT("No DirectDrawClipper object is attached to the surface object."); break;
	case DDERR_NOCOLORCONVHW:
		return TEXT("No color-conversion hardware is present or available."); break;
	case DDERR_NOCOLORKEY:
		return TEXT("The surface does not currently have a color key."); break;
	case DDERR_NOCOLORKEYHW:
		return TEXT("There is no hardware support for the destination color key."); break;
	case DDERR_NOCOOPERATIVELEVELSET:
		return TEXT("A create function was called without the IDirectDraw7::SetCooperativeLevel method."); break;
	case DDERR_NODC:
		return TEXT("No device context (DC) has ever been created for this surface."); break;
	case DDERR_NODDROPSHW:
		return TEXT("No DirectDraw raster-operation (ROP) hardware is available."); break;
	case DDERR_NODIRECTDRAWHW:
		return TEXT("Hardware-only DirectDraw object creation is not possible; break; the driver does not support any hardware."); break;
	case DDERR_NODIRECTDRAWSUPPORT:
		return TEXT("DirectDraw support is not possible with the current display driver."); break;
	case DDERR_NOEMULATION:
		return TEXT("Software emulation is not available."); break;
	case DDERR_NOEXCLUSIVEMODE:
		return TEXT("The operation requires the application to have exclusive mode, but the application does not have exclusive mode."); break;
	case DDERR_NOFLIPHW:
		return TEXT("Flipping visible surfaces is not supported."); break;
	case DDERR_NOFOCUSWINDOW:
		return TEXT("An attempt was made to create or set a device window without first setting the focus window."); break;
	case DDERR_NOGDI:
		return TEXT("No GDI is present."); break;
	case DDERR_NOHWND:
		return TEXT("Clipper notification requires a window handle, or no window handle has been previously set as the cooperative level window handle."); break;
	case DDERR_NOMIPMAPHW:
		return TEXT("No mipmap-capable texture mapping hardware is present or available."); break;
	case DDERR_NOMIRRORHW:
		return TEXT("No mirroring hardware is present or available."); break;
	case DDERR_NONONLOCALVIDMEM:
		return TEXT("An attempt was made to allocate nonlocal video memory from a device that does not support nonlocal video memory."); break;
	case DDERR_NOOPTIMIZEHW:
		return TEXT("The device does not support optimized surfaces."); break;
	case DDERR_NOOVERLAYDEST:
		return TEXT("The IDirectDrawSurface7::GetOverlayPosition method is called on an overlay that the IDirectDrawSurface7::UpdateOverlay method has not been called on to establish as a destination."); break;
	case DDERR_NOOVERLAYHW:
		return TEXT("No overlay hardware is present or available."); break;
	case DDERR_NOPALETTEATTACHED:
		return TEXT("No palette object is attached to this surface."); break;
	case DDERR_NOPALETTEHW:
		return TEXT("There is no hardware support for 16- or 256-color palettes."); break;
	case DDERR_NORASTEROPHW:
		return TEXT("No appropriate raster-operation hardware is present or available."); break;
	case DDERR_NOROTATIONHW:
		return TEXT("No rotation hardware is present or available."); break;
	case DDERR_NOSTRETCHHW:
		return TEXT("There is no hardware support for stretching."); break;
	case DDERR_NOT4BITCOLOR:
		return TEXT("The DirectDrawSurface object is not using a 4-bit color palette, and the requested operation requires a 4-bit color palette."); break;
	case DDERR_NOT4BITCOLORINDEX:
		return TEXT("The DirectDrawSurface object is not using a 4-bit color index palette, and the requested operation requires a 4-bit color index palette."); break;
	case DDERR_NOT8BITCOLOR:
		return TEXT("The DirectDrawSurface object is not using an 8-bit color palette, and the requested operation requires an 8-bit color palette."); break;
	case DDERR_NOTAOVERLAYSURFACE:
		return TEXT("An overlay component is called for a nonoverlay surface."); break;
	case DDERR_NOTEXTUREHW:
		return TEXT("The operation cannot be carried out because no texture-mapping hardware is present or available."); break;
	case DDERR_NOTFLIPPABLE:
		return TEXT("An attempt was made to flip a surface that cannot be flipped."); break;
	case DDERR_NOTFOUND:
		return TEXT("The requested item was not found."); break;
	case DDERR_NOTINITIALIZED:
		return TEXT("An attempt was made to call an interface method of a DirectDraw object created by CoCreateInstance before the object was initialized."); break;
	case DDERR_NOTLOADED:
		return TEXT("The surface is an optimized surface, but it has not yet been allocated any memory."); break;
	case DDERR_NOTLOCKED:
		return TEXT("An attempt was made to unlock a surface that was not locked."); break;
	case DDERR_NOTPAGELOCKED:
		return TEXT("An attempt was made to page-unlock a surface with no outstanding page locks."); break;
	case DDERR_NOTPALETTIZED:
		return TEXT("The surface being used is not a palette-based surface."); break;
	case DDERR_NOVSYNCHW:
		return TEXT("There is no hardware support for vertical blanksynchronized operations."); break;
	case DDERR_NOZBUFFERHW:
		return TEXT("The operation to create a z-buffer in display memory or to perform a blit, using a z-buffer cannot be carried out because there is no hardware support for z-buffers."); break;
	case DDERR_NOZOVERLAYHW:
		return TEXT("The overlay surfaces cannot be z-layered, based on the z-order because the hardware does not support z-ordering of overlays."); break;
	case DDERR_OUTOFCAPS:
		return TEXT("The hardware needed for the requested operation has already been allocated."); break;
	case DDERR_OUTOFMEMORY:
		return TEXT("DirectDraw does not have enough memory to perform the operation."); break;
	case DDERR_OUTOFVIDEOMEMORY:
		return TEXT("DirectDraw does not have enough display memory to perform the operation."); break;
	case DDERR_OVERLAPPINGRECTS:
		return TEXT("The source and destination rectangles are on the same surface and overlap each other."); break;
	case DDERR_OVERLAYCANTCLIP:
		return TEXT("The hardware does not support clipped overlays."); break;
	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE:
		return TEXT("An attempt was made to have more than one color key active on an overlay."); break;
	case DDERR_OVERLAYNOTVISIBLE:
		return TEXT("The IDirectDrawSurface7::GetOverlayPosition method was called on a hicase DDen overlay."); break;
	case DDERR_PALETTEBUSY:
		return TEXT("Access to this palette is refused because the palette is locked by another thread."); break;
	case DDERR_PRIMARYSURFACEALREADYEXISTS:
		return TEXT("This process has already created a primary surface."); break;
	case DDERR_REGIONTOOSMALL:
		return TEXT("The region passed to the IDirectDrawClipper::GetClipList method is too small."); break;
	case DDERR_SURFACEALREADYATTACHED:
		return TEXT("An attempt was made to attach a surface to another surface to which it is already attached."); break;
	case DDERR_SURFACEALREADYDEPENDENT:
		return TEXT("An attempt was made to make a surface a dependency of another surface on which it is already dependent."); break;
	case DDERR_SURFACEBUSY:
		return TEXT("Access to the surface is refused because the surface is locked by another thread."); break;
	case DDERR_SURFACEISOBSCURED:
		return TEXT("Access to the surface is refused because the surface is obscured."); break;
	case DDERR_SURFACELOST:
		return TEXT("Access to the surface is refused because the surface memory is gone. Call the IDirectDrawSurface7::Restore method on this surface to restore the memory associated with it."); break;
	case DDERR_SURFACENOTATTACHED:
		return TEXT("The requested surface is not attached."); break;
	case DDERR_TOOBIGHEIGHT:
		return TEXT("The height requested by DirectDraw is too large."); break;
	case DDERR_TOOBIGSIZE:
		return TEXT("The size requested by DirectDraw is too large. However, the individual height and width are valid sizes."); break;
	case DDERR_TOOBIGWIDTH:
		return TEXT("The width requested by DirectDraw is too large."); break;
	case DDERR_UNSUPPORTED:
		return TEXT("The operation is not supported."); break;
	case DDERR_UNSUPPORTEDFORMAT:
		return TEXT("The pixel format requested is not supported by DirectDraw."); break;
	case DDERR_UNSUPPORTEDMASK:
		return TEXT("The bitmask in the pixel format requested is not supported by DirectDraw."); break;
	case DDERR_UNSUPPORTEDMODE:
		return TEXT("The display is currently in an unsupported mode."); break;
	case DDERR_VERTICALBLANKINPROGRESS:
		return TEXT("A vertical blank is in progress."); break;
	case DDERR_VIDEONOTACTIVE:
		return TEXT("The video port is not active."); break;
	case DDERR_WASSTILLDRAWING:
		return TEXT("The previous blit operation that is transferring information to or from this surface is incomplete."); break;
	case DDERR_WRONGMODE:
		return TEXT("This surface cannot be restored because it was created in a different mode."); break;
	case DDERR_XALIGN:
		return TEXT("The provided rectangle was not horizontally aligned on a required boundary"); break;

	case D3DERR_BADMAJORVERSION:
		return TEXT("The service that you requested is unavailable in this major version of DirectX. (A major version denotes a primary release, such as DirectX 6.0.)");
	case D3DERR_BADMINORVERSION:
		return TEXT("The service that you requested is available in this major version of DirectX, but not in this minor version. Get the latest version of the component run time from Microsoft. (A minor version denotes a secondary release, such as DirectX 6.1.)");
	case D3DERR_COLORKEYATTACHED:
		return TEXT("The application attempted to create a texture with a surface that uses a color key for transparency.");
	case D3DERR_CONFLICTINGTEXTUREFILTER:
		return TEXT("The current texture filters cannot be used together.");
	case D3DERR_CONFLICTINGTEXTUREPALETTE:
		return TEXT("The current textures cannot be used simultaneously. This generally occurs when a multitexture device requires that all palettized textures simultaneously enabled also share the same palette.");
	case D3DERR_CONFLICTINGRENDERSTATE:
		return TEXT("The currently set render states cannot be used together.");
	case D3DERR_DEVICEAGGREGATED:
		return TEXT("The IDirect3DDevice7::SetRenderTarget method was called on a device that was retrieved from the render target surface.");
	case D3DERR_EXECUTE_CLIPPED_FAILED:
		return TEXT("The execute buffer could not be clipped during execution.");
	case D3DERR_EXECUTE_CREATE_FAILED:
		return TEXT("The execute buffer could not be created. This typically occurs when no memory is available to allocate the execute buffer.");
	case D3DERR_EXECUTE_DESTROY_FAILED:
		return TEXT("The memory for the execute buffer could not be deallocated.");
	case D3DERR_EXECUTE_FAILED:
		return TEXT("The contents of the execute buffer are invalid and cannot be executed.");
	case D3DERR_EXECUTE_LOCK_FAILED:
		return TEXT("The execute buffer could not be locked.");
	case D3DERR_EXECUTE_LOCKED:
		return TEXT("The operation requested by the application could not be completed because the execute buffer is locked.");
	case D3DERR_EXECUTE_NOT_LOCKED:
		return TEXT("The execute buffer could not be unlocked because it is not currently locked.");
	case D3DERR_EXECUTE_UNLOCK_FAILED:
		return TEXT("The execute buffer could not be unlocked.");
	case D3DERR_INBEGIN:
		return TEXT("The requested operation cannot be completed while scene rendering is taking place. Try again after the scene is completed and the IDirect3DDevice7::EndScene method is called.");
	case D3DERR_INITFAILED:
		return TEXT("A rendering device could not be created because the new device could not be initialized.");
	case D3DERR_INVALID_DEVICE:
		return TEXT("The requested device type is not valid.");
	case D3DERR_INVALIDCURRENTVIEWPORT:
		return TEXT("The currently selected viewport is not valid.");
	case D3DERR_INVALIDMATRIX:
		return TEXT("The requested operation could not be completed because the combination of the currently set world, view, and projection matrices is invalid (the determinant of the combined matrix is 0).");
	case D3DERR_INVALIDPALETTE:
		return TEXT("The palette associated with a surface is invalid.");
	case D3DERR_INVALIDPRIMITIVETYPE:
		return TEXT("The primitive type specified by the application is invalid.");
	case D3DERR_INVALIDRAMPTEXTURE:
		return TEXT("Ramp mode is being used, and the texture handle in the current material does not match the current texture handle that is set as a render state.");
	case D3DERR_INVALIDVERTEXFORMAT:
		return TEXT("The combination of flexible vertex format flags specified by the application is not valid.");
	case D3DERR_INVALIDVERTEXTYPE:
		return TEXT("The vertex type specified by the application is invalid.");
	case D3DERR_LIGHT_SET_FAILED:
		return TEXT("The attempt to set lighting parameters for a light object failed.");
	case D3DERR_LIGHTHASVIEWPORT:
		return TEXT("The requested operation failed because the light object is associated with another viewport.");
	case D3DERR_LIGHTNOTINTHISVIEWPORT:
		return TEXT("The requested operation failed because the light object has not been associated with this viewport.");
	case D3DERR_MATERIAL_CREATE_FAILED:
		return TEXT("The material could not be created. This typically occurs when no memory is available to allocate for the material.");
	case D3DERR_MATERIAL_DESTROY_FAILED:
		return TEXT("The memory for the material could not be deallocated.");
	case D3DERR_MATERIAL_GETDATA_FAILED:
		return TEXT("The material parameters could not be retrieved.");
	case D3DERR_MATERIAL_SETDATA_FAILED:
		return TEXT("The material parameters could not be set.");
	case D3DERR_MATRIX_CREATE_FAILED:
		return TEXT("The matrix could not be created. This can occur when no memory is available to allocate for the matrix.");
	case D3DERR_MATRIX_DESTROY_FAILED:
		return TEXT("The memory for the matrix could not be deallocated.");
	case D3DERR_MATRIX_GETDATA_FAILED:
		return TEXT("The matrix data could not be retrieved. This can occur when the matrix was not created by the current device.");
	case D3DERR_MATRIX_SETDATA_FAILED:
		return TEXT("The matrix data could not be set. This can occur when the matrix was not created by the current device.");
	case D3DERR_NOCURRENTVIEWPORT:
		return TEXT("The viewport parameters could not be retrieved because none have been set.");
	case D3DERR_NOTINBEGIN:
		return TEXT("The requested rendering operation could not be completed because scene rendering has not begun. Call IDirect3DDevice7::BeginScene to begin rendering, and try again.");
	case D3DERR_NOVIEWPORTS:
		return TEXT("The requested operation failed because the device currently has no viewports associated with it.");
	case D3DERR_SCENE_BEGIN_FAILED:
		return TEXT("Scene rendering could not begin.");
	case D3DERR_SCENE_END_FAILED:
		return TEXT("Scene rendering could not be completed.");
	case D3DERR_SCENE_IN_SCENE:
		return TEXT("Scene rendering could not begin because a previous scene was not completed by a call to the IDirect3DDevice7::EndScene method.");
	case D3DERR_SCENE_NOT_IN_SCENE:
		return TEXT("Scene rendering could not be completed because a scene was not started by a previous call to the IDirect3DDevice7::BeginScene method.");
	case D3DERR_SETVIEWPORTDATA_FAILED:
		return TEXT("The viewport parameters could not be set.");
	case D3DERR_STENCILBUFFER_NOTPRESENT:
		return TEXT("The requested stencil buffer operation could not be completed because there is no stencil buffer attached to the render target surface.");
	case D3DERR_SURFACENOTINVIDMEM:
		return TEXT("The device could not be created because the render target surface is not located in video memory. (Hardware-accelerated devices require video-memory render target surfaces.)");
	case D3DERR_TEXTURE_BADSIZE:
		return TEXT("The dimensions of a current texture are invalid. This can occur when an application attempts to use a texture that has dimensions that are not a power of 2 with a device that requires them.");
	case D3DERR_TEXTURE_CREATE_FAILED:
		return TEXT("The texture handle for the texture could not be retrieved from the driver.");
	case D3DERR_TEXTURE_DESTROY_FAILED:
		return TEXT("The device was unable to deallocate the texture memory.");
	case D3DERR_TEXTURE_GETSURF_FAILED:
		return TEXT("The DirectDraw surface used to create the texture could not be retrieved.");
	case D3DERR_TEXTURE_LOAD_FAILED:
		return TEXT("The texture could not be loaded.");
	case D3DERR_TEXTURE_LOCK_FAILED:
		return TEXT("The texture could not be locked.");
	case D3DERR_TEXTURE_LOCKED:
		return TEXT("The requested operation could not be completed because the texture surface is currently locked.");
	case D3DERR_TEXTURE_NO_SUPPORT:
		return TEXT("The device does not support texture mapping.");
	case D3DERR_TEXTURE_NOT_LOCKED:
		return TEXT("The requested operation could not be completed because the texture surface is not locked.");
	case D3DERR_TEXTURE_SWAP_FAILED:
		return TEXT("The texture handles could not be swapped.");
	case D3DERR_TEXTURE_UNLOCK_FAILED:
		return TEXT("The texture surface could not be unlocked.");
	case D3DERR_TOOMANYOPERATIONS:
		return TEXT("The application is requesting more texture-filtering operations than the device supports.");
	case D3DERR_TOOMANYPRIMITIVES:
		return TEXT("The device is unable to render the provided number of primitives in a single pass.");
	case D3DERR_UNSUPPORTEDALPHAARG:
		return TEXT("The device does not support one of the specified texture-blending arguments for the alpha channel.");
	case D3DERR_UNSUPPORTEDALPHAOPERATION:
		return TEXT("The device does not support one of the specified texture-blending operations for the alpha channel.");
	case D3DERR_UNSUPPORTEDCOLORARG:
		return TEXT("The device does not support one of the specified texture-blending arguments for color values.");
	case D3DERR_UNSUPPORTEDCOLOROPERATION:
		return TEXT("The device does not support one of the specified texture-blending operations for color values.");
	case D3DERR_UNSUPPORTEDFACTORVALUE:
		return TEXT("The specified texture factor value is not supported by the device.");
	case D3DERR_UNSUPPORTEDTEXTUREFILTER:
		return TEXT("The specified texture filter is not supported by the device.");
	case D3DERR_VBUF_CREATE_FAILED:
		return TEXT("The vertex buffer could not be created. This can happen when there is insufficient memory to allocate a vertex buffer.");
	case D3DERR_VERTEXBUFFERLOCKED:
		return TEXT("The requested operation could not be completed because the vertex buffer is locked.");
	case D3DERR_VERTEXBUFFEROPTIMIZED:
		return TEXT("The requested operation could not be completed because the vertex buffer is optimized. (The contents of optimized vertex buffers are driver-specific and considered private.)");
	case D3DERR_VIEWPORTDATANOTSET:
		return TEXT("The requested operation could not be completed because viewport parameters have not yet been set. Set the viewport parameters by calling the IDirect3DDevice7::SetViewport method, and try again.");
	case D3DERR_VIEWPORTHASNODEVICE:
		return TEXT("This value is used only by the IDirect3DDevice3 interface and its predecessors. For the IDirect3DDevice7 interface, this error value is not used. The requested operation could not be completed because the viewport has not yet been associated with a device. Associate the viewport with a rendering device by calling the IDirect3DDevice3::AddViewport method, and try again.");
	case D3DERR_WRONGTEXTUREFORMAT:
		return TEXT("The pixel format of the texture surface is not valid.");
	case D3DERR_ZBUFF_NEEDS_SYSTEMMEMORY:
		return TEXT("The requested operation could not be completed because the specified device requires system-memory depth-buffer surfaces. (Software rendering devices require system-memory depth buffers.)");
	case D3DERR_ZBUFF_NEEDS_VIDEOMEMORY:
		return TEXT("The requested operation could not be completed because the specified device requires video-memory depth-buffer surfaces. (Hardware-accelerated devices require video-memory depth buffers.)");
	case D3DERR_ZBUFFER_NOTPRESENT:
		return TEXT("The requested operation could not be completed because the render target surface does not have an attached depth buffer.");
		
	default:
		return TEXT("Unknown Error");
	}
}

static HRESULT InitFail(HWND hWnd, HRESULT hr, LPCTSTR error) {
    TCHAR szBuff[128];

    wsprintf(szBuff, TEXT("%s\n%s"), error, ErrorMsg(hr));
    DestroyAll();
    OutputDebugString(szBuff);
    
	DWORD tick;
	tick = GetTickCount();							// Set the timer
	while(GetTickCount() < tick + 3000);		// Wait
    return hr;
}

// ---------------------------------------------------------------------------------------- //
//  WindowProc()                                                                            //
//  Does very little more than allow us to kill the App by hitting ALT+F4 under PC Windows  //
//  Might even work on the Dreamcast, who knows...                                          //
//  Timer code is left here in case we need it.                                             //
//  I'm not sure we'll ever use this function in the actual Dreamcast code.                 //
// ---------------------------------------------------------------------------------------- //
long FAR PASCAL WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	const int numpics = 12;
	const int picres[numpics] = { IDB_BREEZE1, IDB_BREEZE2, IDB_BREEZE3, IDB_BREEZE4, 
		IDB_PIXELDOTZ1, IDB_PIXELDOTZ2, IDB_BURNER0, IDB_MARTIN, IDB_DAVEP1, IDB_DAVEP2,
		IDB_SPLASH, IDB_PFUNK };
	static int picnum = 0;
	static LPDIRECTDRAWSURFACE4 bitmap = NULL;
    
	switch (message) {
        case WM_KILLFOCUS:
            SetFocus(hWnd);
            return 0L;

        case WM_DESTROY:
			DestroyAll();
            PostQuitMessage(0);
            return 0L;

        case WM_TIMER:
			if(bitmap) bitmap->Release();
			bitmap = LoadBitmap(picres[picnum]);
			gdds_back->Blt(NULL, bitmap, NULL, DDBLT_WAIT, NULL);
			gdds_screen->Flip(NULL, DDFLIP_WAIT);
			picnum++;
			if(picnum >= numpics) picnum = 0;
            break;
    }
    return DefWindowProc(hWnd, message, wParam, lParam);
}

void SetDisplayMode(int x, int y, int bpp) {
    DDSURFACEDESC2	ddsd;
	HRESULT hr;

	if(gdds_screen) gdds_screen->Release();
	if(gdds_back) gdds_back->Release();

#if !defined WINDOWED
	hr = gpdd_directdraw->SetDisplayMode(x, y, bpp, 0, 0);
    if (hr != DD_OK) { InitFail(ghwd_window, hr, TEXT("Error Setting Display Mode")); }
#endif

    // Create the primary surface
    memset(&ddsd, 0, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
#if !defined WINDOWED
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX | DDSCAPS_3DDEVICE;
	ddsd.dwBackBufferCount = 1;
#else 
    ddsd.dwFlags = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_3DDEVICE;
	ddsd.dwBackBufferCount = 1;
#endif
    hr = gpdd_directdraw->CreateSurface(&ddsd, &gdds_screen, NULL);

    if(hr != DD_OK) { InitFail(ghwd_window, hr, TEXT("Error Creating Primary surface")); }

	// Get the back buffer
#if !defined WINDOWED
	ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
	hr = gdds_screen->GetAttachedSurface(&ddsd.ddsCaps, &gdds_back);
#else
	gdds_back = gdds_screen;
#endif

	if(hr != DD_OK) { InitFail(ghwd_window, hr, TEXT("Error Obtaining Back surface")); }
}

// ----------------------------------------------------------------- //
//  InitApp()                                                        //
//  Creates the logical windows and initializes DirectDraw using it  //
// ----------------------------------------------------------------- //
static HRESULT InitApp(HINSTANCE hInstance, int nCmdShow) {

    HWND			hwnd;
    WNDCLASS		cls;
    DDSURFACEDESC2	ddsd;
    DDCAPS			ddcaps;
    HRESULT			hr;
    LPDIRECTDRAW	dd;

    // Set up and register window class.

	cls.hCursor        = NULL;
	cls.hIcon          = NULL;
	cls.lpszMenuName   = NULL;
	cls.hbrBackground  = NULL;
	cls.hInstance      = hInstance;
	cls.lpszClassName  = NAME;
	cls.lpfnWndProc    = (WNDPROC)WindowProc;
	cls.style          = 0;
	cls.cbWndExtra     = 0;
	cls.cbClsExtra     = 0;

	if (!RegisterClass(&cls))
		return FALSE;

    // Create a window.

    hwnd = CreateWindowEx(
		0,
		NAME,
		TITLE,
#if !defined USE_DIRECT_X
		WS_OVERLAPPEDWINDOW | WS_VISIBLE,
#elif defined WINDOWED
		WS_OVERLAPPEDWINDOW | WS_VISIBLE,
#else
		WS_VISIBLE | WS_POPUP,
#endif
		0,
		0,
		640,
		480,
		NULL,
		NULL,
		hInstance,
		NULL);

    if (!hwnd)
        return FALSE;
    SetFocus(hwnd);
	ghwd_window = hwnd;

#if defined USE_DIRECT_X
    // Create the DirectDraw object
    hr = DirectDrawCreate(NULL, &dd, NULL);
    if (hr != DD_OK) { InitFail(hwnd, hr, TEXT("Error Creating DirectDraw object")); }
        
    // Get DirectDraw4 interface.
    hr = dd->QueryInterface(IID_IDirectDraw4, (LPVOID *)&gpdd_directdraw);
    if (hr != DD_OK) { InitFail(hwnd, hr, TEXT("Error Creating IDirectDraw4 object")); }
    dd->Release();
    dd = NULL;

	// Full screen exclusive
#if ! defined WINDOWED
    hr = gpdd_directdraw->SetCooperativeLevel(hwnd, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE);
    if (hr != DD_OK) { InitFail(hwnd, hr, TEXT("Error Setting Cooperate level")); }
#else
    hr = gpdd_directdraw->SetCooperativeLevel(hwnd, DDSCL_NORMAL);
    if (hr != DD_OK) { InitFail(hwnd, hr, TEXT("Error Setting Cooperate level")); }
#endif

#if defined MENU_640_480
	SetDisplayMode(640, 480, 16);
#else
	SetDisplayMode(320, 240, 16);
#endif

	InitDirect3D();
	gdev_device = GetDeviceFromSurface(gdds_back, NULL);

#endif

	return DD_OK;

}

// --------------------------------------------------------- //
//  LoadBitmap()                                             //
//  This should load a bitmap into a new DirectDraw surface  //
// --------------------------------------------------------- //

LPDIRECTDRAWSURFACE4 LoadBitmap(int res) {
    HBITMAP					hbitmap;    // Bitmap handle
    BITMAP					bitmap;     // Bitmap information
    DDSURFACEDESC2			ddsd;       // Surface Descriptor
    LPDIRECTDRAWSURFACE4	pdds;       // Pointer to a DirectDraw surface
    HDC						hdcBitmap;  // DC for bitmap
    HDC						hdcSurface; // DC for DDraw surface
	HRESULT					hr;

    // Try to Load the bitmap as a resource or a file
    hbitmap = (HBITMAP)LoadImage (gh_inst, MAKEINTRESOURCE(res), IMAGE_BITMAP, 0, 0, LR_DEFAULTCOLOR);
    if (hbitmap== NULL)
    {
		HDC dc;
		TCHAR str[256];
		int n;

		gdds_screen->GetDC(&dc);
		SetTextColor(dc, RGB(255, 255, 255));
		SetBkMode(dc, TRANSPARENT);

		n = wsprintf(str, TEXT("Bitmap not found"), ddsd.dwWidth, ddsd.dwHeight);
		ExtTextOut(dc, 10, 96, NULL, NULL, str, n, NULL);

		gdds_screen->ReleaseDC(dc);
        return NULL;
    }

    // Get the width, height and other info about the bitmap.
    GetObject (hbitmap, sizeof(BITMAP), &bitmap);

    // Create a DirectDrawSurface for this texture
    memset(&ddsd, 0, sizeof(ddsd));
    gdds_screen->GetSurfaceDesc(&ddsd);  

    ddsd.dwSize         = sizeof(DDSURFACEDESC2);
    ddsd.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_VIDEOMEMORY;

    ddsd.dwWidth        = bitmap.bmWidth;
    ddsd.dwHeight       = bitmap.bmHeight;
    hr = gpdd_directdraw->CreateSurface(&ddsd, &pdds, NULL);
	if(hr != DD_OK) { 
		HDC dc;
		TCHAR str[256];
		int n;

		gdds_screen->GetDC(&dc);
		SetTextColor(dc, RGB(255, 255, 255));
		SetBkMode(dc, TRANSPARENT);

		n = wsprintf(str, TEXT("DirectDraw Surface not Created"), ddsd.dwWidth, ddsd.dwHeight);
		ExtTextOut(dc, 10, 112, NULL, NULL, str, n, NULL);
		n = wsprintf(str, TEXT("%s"), ErrorMsg(hr));
		ExtTextOut(dc, 10, 128, NULL, NULL, str, n, NULL);

		gdds_screen->ReleaseDC(dc);
        return NULL;
	}

    // Create an offscreen DC, and load the bitmap into it.
    hdcBitmap = CreateCompatibleDC (NULL);
    SelectObject (hdcBitmap, hbitmap);

    // Get a DC onto the DDraw surface.
    hr = pdds->GetDC (&hdcSurface);
	if(hr != DD_OK) {
		HDC dc;
		TCHAR str[256];
		int n;

		gdds_screen->GetDC(&dc);
		SetTextColor(dc, RGB(255, 255, 255));
		SetBkMode(dc, TRANSPARENT);

		n = wsprintf(str, TEXT("Could obtain a Device Context"), ddsd.dwWidth, ddsd.dwHeight);
		ExtTextOut(dc, 10, 128, NULL, NULL, str, n, NULL);

		gdds_screen->ReleaseDC(dc);
        return NULL;
	}

    // Copy the bits from the bitmap to the DDraw surface.
    BitBlt (hdcSurface, 0, 0, bitmap.bmWidth, bitmap.bmHeight, hdcBitmap, 0, 0, SRCCOPY);

    // Clean up after ourselves.
    pdds->ReleaseDC (hdcSurface);
    DeleteDC (hdcBitmap);
    DeleteObject (hbitmap);

    return pdds;
}

// ------------ //
//  WinMain()   //
//  What else?  //
//------------- //
int PASCAL WinMain(
		HINSTANCE hInstance,
		HINSTANCE, 
#ifdef UNDER_CE
        LPWSTR,
#else
        LPSTR,
#endif
        int nCmdShow) {

    MSG msg;

    gh_inst = hInstance;

    if(InitApp(hInstance, nCmdShow) != DD_OK)
        return FALSE;

	DoGUI();

/*	// This code will not stay here
	// Here I will obtain some information about the DirectDraw object and display it
	// Using GDI text surfaces and the default font.
    DDSURFACEDESC2	ddsd;
    DDCAPS			ddcaps;
	HDC				dc;
	TCHAR			str[1024];
	int n;

	ddsd.dwSize = sizeof(ddsd);
	gdds_screen->GetSurfaceDesc(&ddsd);
	gdds_screen->GetDC(&dc);
	SetTextColor(dc, RGB(255, 255, 255));
	SetBkMode(dc, TRANSPARENT);

	n = wsprintf(str, TEXT("Surface Dimensions: (%i x %i)"), ddsd.dwWidth, ddsd.dwHeight);
	ExtTextOut(dc, 0, 0, NULL, NULL, str, n, NULL);
	n = wsprintf(str, TEXT("Pitch: %i"), ddsd.lPitch);
	ExtTextOut(dc, 0, 16, NULL, NULL, str, n, NULL);

	n = wsprintf(str, TEXT("Bits per Pixel: %i"), ddsd.ddpfPixelFormat.dwRGBBitCount);
	ExtTextOut(dc, 0, 32, NULL, NULL, str, n, NULL);
	n = wsprintf(str, TEXT("Bit Masks: Red - 0x%X, Green - 0x%X,"),
		ddsd.ddpfPixelFormat.dwRBitMask,
		ddsd.ddpfPixelFormat.dwGBitMask);
	ExtTextOut(dc, 0, 48, NULL, NULL, str, n, NULL);
	n = wsprintf(str, TEXT("   Blue - 0x%X"), ddsd.ddpfPixelFormat.dwBBitMask);
	ExtTextOut(dc, 0, 64, NULL, NULL, str, n, NULL);



	gdds_screen->ReleaseDC(dc); */

	// More code that doesn't belong here

/*	RECT rt = { 0, 240, 320, 480 };
	LPDIRECTDRAWSURFACE4 bitmap = LoadBitmap(IDB_BITMAP1);
	if(bitmap == NULL) {
		gdds_screen->GetDC(&dc);
		SetTextColor(dc, RGB(255, 255, 255));
		SetBkMode(dc, TRANSPARENT);

		n = wsprintf(str, TEXT("Unknown Error Loading Bitmap"), ddsd.dwWidth, ddsd.dwHeight);
		ExtTextOut(dc, 0, 80, NULL, NULL, str, n, NULL);

		gdds_screen->ReleaseDC(dc);
	} else {
		HRESULT hr = gdds_back->Blt(NULL, bitmap, NULL, DDBLT_WAIT, NULL);
		if(hr != DD_OK) {
			gdds_screen->GetDC(&dc);
			SetTextColor(dc, RGB(255, 255, 255));
			SetBkMode(dc, TRANSPARENT);

			n = wsprintf(str, TEXT("Error Displaying Bitmap"), ddsd.dwWidth, ddsd.dwHeight);
			ExtTextOut(dc, 0, 144, NULL, NULL, str, n, NULL);
			n = wsprintf(str, TEXT("%s"), ErrorMsg(hr));
			ExtTextOut(dc, 10, 160, NULL, NULL, str, n, NULL);

			gdds_screen->ReleaseDC(dc);
		} else {
			gdds_screen->GetDC(&dc);
			SetTextColor(dc, RGB(255, 255, 255));
			SetBkMode(dc, TRANSPARENT);

			n = wsprintf(str, TEXT("HRESULT == DD_OK"), ddsd.dwWidth, ddsd.dwHeight);
			ExtTextOut(dc, 0, 144, NULL, NULL, str, n, NULL);
			n = wsprintf(str, TEXT("%s"), ErrorMsg(hr));
			ExtTextOut(dc, 10, 160, NULL, NULL, str, n, NULL);

			gdds_screen->ReleaseDC(dc);
		}
	}

	gdds_screen->Flip(NULL, DDFLIP_WAIT); */



	// End of code that doesn't belong here

//	SetTimer(ghwd_window, 1, 3000, NULL);


//    while(GetMessage(&msg, NULL, 0, 0)) {
//        TranslateMessage(&msg);
//        DispatchMessage(&msg);
//    }
    return 0;
}
